/*
 * zaphfc.c - Zaptel driver for HFC-S PCI A based ISDN BRI cards
 *
 * kernel module inspired by HFC PCI ISDN4Linux and Zaptel drivers
 *
 * Copyright (C) 2002, 2003, 2004, 2005 Junghanns.NET GmbH
 *
 * Klaus-Peter Junghanns <kpj@junghanns.net>
 *
 * Copyright (C) 2004, 2005  Florian Zumbiehl <florz@gmx.de>
 *  - support for slave mode of the HFC-S chip which allows it to
 *    sync its sample clock to an external source/another HFC chip
 *  - support for "interrupt bundling" (let only one card generate
 *    8 kHz timing interrupt no matter how many cards there are
 *    in the system)
 *  - interrupt loss tolerant b channel handling
 *
 * This program is free software and may be modified and
 * distributed under the terms of the GNU General Public License.
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <zaptel.h>
#include "zaphfc.h"

#define log2(n) ffz(~(n))

#if CONFIG_PCI

#define CLKDEL_TE	0x0f	/* CLKDEL in TE mode */
#define CLKDEL_NT	0x6c	/* CLKDEL in NT mode */

typedef struct {
        int vendor_id;
        int device_id;
        char *vendor_name;
        char *card_name;
} PCI_ENTRY;

static const PCI_ENTRY id_list[] =
{
        {PCI_VENDOR_ID_CCD, PCI_DEVICE_ID_CCD_2BD0, "CCD/Billion/Asuscom", "2BD0"},
        {PCI_VENDOR_ID_CCD, PCI_DEVICE_ID_CCD_B000, "Billion", "B000"},
        {PCI_VENDOR_ID_CCD, PCI_DEVICE_ID_CCD_B006, "Billion", "B006"},
        {PCI_VENDOR_ID_CCD, PCI_DEVICE_ID_CCD_B007, "Billion", "B007"},
        {PCI_VENDOR_ID_CCD, PCI_DEVICE_ID_CCD_B008, "Billion", "B008"},
        {PCI_VENDOR_ID_CCD, PCI_DEVICE_ID_CCD_B009, "Billion", "B009"},
        {PCI_VENDOR_ID_CCD, PCI_DEVICE_ID_CCD_B00A, "Billion", "B00A"},
        {PCI_VENDOR_ID_CCD, PCI_DEVICE_ID_CCD_B00B, "Billion", "B00B"},
        {PCI_VENDOR_ID_CCD, PCI_DEVICE_ID_CCD_B00C, "Billion", "B00C"},
        {PCI_VENDOR_ID_CCD, PCI_DEVICE_ID_CCD_B100, "Seyeon", "B100"},
        {PCI_VENDOR_ID_ABOCOM, PCI_DEVICE_ID_ABOCOM_2BD1, "Abocom/Magitek", "2BD1"},
        {PCI_VENDOR_ID_ASUSTEK, PCI_DEVICE_ID_ASUSTEK_0675, "Asuscom/Askey", "675"},
        {PCI_VENDOR_ID_BERKOM, PCI_DEVICE_ID_BERKOM_T_CONCEPT, "German telekom", "T-Concept"},
        {PCI_VENDOR_ID_BERKOM, PCI_DEVICE_ID_BERKOM_A1T, "German telekom", "A1T"},
        {PCI_VENDOR_ID_ANIGMA, PCI_DEVICE_ID_ANIGMA_MC145575, "Motorola MC145575", "MC145575"},
        {PCI_VENDOR_ID_ZOLTRIX, PCI_DEVICE_ID_ZOLTRIX_2BD0, "Zoltrix", "2BD0"},
        {PCI_VENDOR_ID_DIGI, PCI_DEVICE_ID_DIGI_DF_M_IOM2_E,"Digi International", "Digi DataFire Micro V IOM2 (Europe)"},
        {PCI_VENDOR_ID_DIGI, PCI_DEVICE_ID_DIGI_DF_M_E,"Digi International", "Digi DataFire Micro V (Europe)"},
        {PCI_VENDOR_ID_DIGI, PCI_DEVICE_ID_DIGI_DF_M_IOM2_A,"Digi International", "Digi DataFire Micro V IOM2 (North America)"},
        {PCI_VENDOR_ID_DIGI, PCI_DEVICE_ID_DIGI_DF_M_A,"Digi International", "Digi DataFire Micro V (North America)"},
	{0x182d, 0x3069,"Sitecom","Isdn 128 PCI"},
        {0, 0, NULL, NULL},
};

static struct hfc_card *hfc_dev_list = NULL;
static int hfc_dev_count = 0;
static int modes = 0; // all TE
static int debug = 0;
static int sync_slave = 0; // all master
static int timer_card = 0;
static struct pci_dev *multi_hfc = NULL;
static spinlock_t registerlock = SPIN_LOCK_UNLOCKED;

void hfc_shutdownCard1(struct hfc_card *hfctmp) {
    printk(KERN_INFO "zaphfc: shutting down card at %p.\n",hfctmp->pci_io);

    /* Clear interrupt mask */
    hfctmp->regs.int_m2 = 0;
    hfc_outb(hfctmp, hfc_INT_M2, hfctmp->regs.int_m2);

    /* Remove interrupt handler */
    free_irq(hfctmp->irq,hfctmp);
}

void hfc_shutdownCard2(struct hfc_card *hfctmp) {
    unsigned long flags;

    spin_lock_irqsave(&hfctmp->lock,flags);

    /* Reset pending interrupts */
    hfc_inb(hfctmp, hfc_INT_S1);

    /* Soft-reset the card */
    hfc_outb(hfctmp, hfc_CIRM, hfc_CIRM_RESET); // softreset on

    spin_unlock_irqrestore(&hfctmp->lock, flags);
    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout((30 * HZ) / 1000);	// wait 30 ms
    spin_lock_irqsave(&hfctmp->lock,flags);

    hfc_outb(hfctmp,hfc_CIRM,0);	// softreset off

    pci_write_config_word(hfctmp->pcidev, PCI_COMMAND, 0);	// disable memio and bustmaster

    if (hfctmp->fifos != NULL) {
	free_pages((unsigned long)hfctmp->fifos,log2(hfc_FIFO_MEM_SIZE_PAGES));
    }
    iounmap((void *) hfctmp->pci_io);
    hfctmp->pci_io = NULL;
    if (hfctmp->pcidev != NULL) {
        pci_disable_device(hfctmp->pcidev);
    }
    spin_unlock_irqrestore(&hfctmp->lock,flags);
    if (hfctmp->ztdev != NULL) {
	zt_unregister(&hfctmp->ztdev->span);
	vfree(hfctmp->ztdev);
	printk(KERN_INFO "unregistered from zaptel.\n");
    }
}

void hfc_shutdownCard(struct hfc_card *hfctmp) {
    if (hfctmp == NULL) {
	return;
    }

    if (hfctmp->pci_io == NULL) {
	return;
    }

    hfc_shutdownCard1(hfctmp);
    hfc_shutdownCard2(hfctmp);
}

void hfc_resetCard(struct hfc_card *hfctmp) {
    unsigned long flags;

    spin_lock_irqsave(&hfctmp->lock,flags);
    pci_write_config_word(hfctmp->pcidev, PCI_COMMAND, PCI_COMMAND_MEMORY);	// enable memio
    hfctmp->regs.int_m2 = 0;
    hfc_outb(hfctmp, hfc_INT_M2, hfctmp->regs.int_m2);

//    printk(KERN_INFO "zaphfc: resetting card.\n");
    pci_set_master(hfctmp->pcidev);
    hfc_outb(hfctmp, hfc_CIRM, hfc_CIRM_RESET);	// softreset on
    spin_unlock_irqrestore(&hfctmp->lock, flags);

    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout((30 * HZ) / 1000);	// wait 30 ms
    hfc_outb(hfctmp, hfc_CIRM, 0);	// softreset off

    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout((20 * HZ) / 1000);	// wait 20 ms
    if (hfc_inb(hfctmp,hfc_STATUS) & hfc_STATUS_PCI_PROC) {
	printk(KERN_WARNING "zaphfc: hfc busy.\n");
    }

//    hfctmp->regs.fifo_en = hfc_FIFOEN_D | hfc_FIFOEN_B1 | hfc_FIFOEN_B2;
//    hfctmp->regs.fifo_en = hfc_FIFOEN_D;	/* only D fifos enabled */
    hfctmp->regs.fifo_en = 0;	/* no fifos enabled */
    hfc_outb(hfctmp, hfc_FIFO_EN, hfctmp->regs.fifo_en);

    hfctmp->regs.trm = 2;
    hfc_outb(hfctmp, hfc_TRM, hfctmp->regs.trm);

    if (hfctmp->regs.nt_mode == 1) {
	hfc_outb(hfctmp, hfc_CLKDEL, CLKDEL_NT); /* ST-Bit delay for NT-Mode */
    } else {
	hfc_outb(hfctmp, hfc_CLKDEL, CLKDEL_TE); /* ST-Bit delay for TE-Mode */
    }
    hfctmp->regs.sctrl_e = hfc_SCTRL_E_AUTO_AWAKE;
    hfc_outb(hfctmp, hfc_SCTRL_E, hfctmp->regs.sctrl_e);	/* S/T Auto awake */
    hfctmp->regs.bswapped = 0;	/* no exchange */

    hfctmp->regs.ctmt = hfc_CTMT_TRANSB1 | hfc_CTMT_TRANSB2; // all bchans are transparent , no freaking hdlc
    hfc_outb(hfctmp, hfc_CTMT, hfctmp->regs.ctmt);

    hfctmp->regs.int_m1=hfc_INTS_L1STATE;
    if(hfctmp->cardno==timer_card){
	hfctmp->regs.int_m2=hfc_M2_PROC_TRANS;
    }else{
	hfctmp->regs.int_m1|=hfc_INTS_DREC;
	hfctmp->regs.int_m2=0;
    }
    hfc_outb(hfctmp, hfc_INT_M1, hfctmp->regs.int_m1);
    hfc_outb(hfctmp, hfc_INT_M2, hfctmp->regs.int_m2);

    /* Clear already pending ints */
    hfc_inb(hfctmp, hfc_INT_S1);

    if (hfctmp->regs.nt_mode == 1) {
	hfctmp->regs.sctrl = 3 | hfc_SCTRL_NONE_CAP | hfc_SCTRL_MODE_NT;	/* set tx_lo mode, error in datasheet ! */
    } else {
	hfctmp->regs.sctrl = 3 | hfc_SCTRL_NONE_CAP | hfc_SCTRL_MODE_TE;	/* set tx_lo mode, error in datasheet ! */
    }

    hfc_outb(hfctmp, hfc_MST_MODE, hfctmp->regs.mst_mode);
    hfc_outb(hfctmp, hfc_MST_EMOD, hfctmp->regs.mst_emod);

    hfc_outb(hfctmp, hfc_SCTRL, hfctmp->regs.sctrl);
    hfctmp->regs.sctrl_r = 3;
    hfc_outb(hfctmp, hfc_SCTRL_R, hfctmp->regs.sctrl_r);

    hfctmp->regs.connect = 0;
    hfc_outb(hfctmp, hfc_CONNECT, hfctmp->regs.connect);

    hfc_outb(hfctmp, hfc_CIRM, 0x80 | 0x40);	// bit order

    /* Finally enable IRQ output */
    hfctmp->regs.int_m2 |= hfc_M2_IRQ_ENABLE;
    hfc_outb(hfctmp, hfc_INT_M2, hfctmp->regs.int_m2);


    hfc_inb(hfctmp, hfc_INT_S1); 
    hfc_inb(hfctmp, hfc_INT_S2);
}

void hfc_registerCard(struct hfc_card *hfccard) {
    spin_lock(&registerlock);
    if (hfccard != NULL) {
	hfccard->cardno = hfc_dev_count++;
	hfccard->next = hfc_dev_list;
	hfc_dev_list = hfccard;
    }
    spin_unlock(&registerlock);
}

/*===========================================================================*/

#if hfc_B_FIFO_SIZE%ZT_CHUNKSIZE
#error hfc_B_FIFO_SIZE is not a multiple of ZT_CHUNKSIZE even though the code assumes this
#endif
    
static void hfc_dch_init(struct hfc_card *hfctmp){
    struct dch *chtmp=&hfctmp->dch;

    chtmp->rx.f1.p=(u8 *)(hfctmp->fifos+hfc_FIFO_DRX_F1);
    chtmp->rx.f2.v=0x1f;
    chtmp->rx.f2.z2.v=0x1ff;

    chtmp->tx.f1.p=(u8 *)(hfctmp->fifos+hfc_FIFO_DTX_F1);
    chtmp->tx.f1.v=0x1f;
    chtmp->tx.f1.z1.v=0x1ff;
    chtmp->tx.f2.p=(u8 *)(hfctmp->fifos+hfc_FIFO_DTX_F2);
}

static void hfc_bch_init(struct hfc_card *hfctmp){
    struct bch *chtmp=&hfctmp->bch;

    chtmp->checkcnt=0;
    chtmp->fill_fifo=0;

    chtmp->rx.c[0].z1p=(unsigned short *)(hfctmp->fifos+hfc_FIFO_B1RX_Z1+0x1f*4);
    chtmp->rx.c[0].fifo_base=(char *)(hfctmp->fifos+hfc_FIFO_B1RX_ZOFF);
    chtmp->rx.c[1].z1p=(unsigned short *)(hfctmp->fifos+hfc_FIFO_B2RX_Z1+0x1f*4);
    chtmp->rx.c[1].fifo_base=(char *)(hfctmp->fifos+hfc_FIFO_B2RX_ZOFF);
    chtmp->rx.z2=hfc_B_SUB_VAL;
    chtmp->rx.diff=0;

    chtmp->tx.c[0].z1p=(unsigned short *)(hfctmp->fifos+hfc_FIFO_B1TX_Z1+0x1f*4);
    chtmp->tx.c[0].z2p=(unsigned short *)(hfctmp->fifos+hfc_FIFO_B1TX_Z2+0x1f*4);
    chtmp->tx.c[0].fifo_base=(char *)(hfctmp->fifos+hfc_FIFO_B1TX_ZOFF);
    chtmp->tx.c[0].filled=0;
    chtmp->tx.c[1].z1p=(unsigned short *)(hfctmp->fifos+hfc_FIFO_B2TX_Z1+0x1f*4);
    chtmp->tx.c[1].z2p=(unsigned short *)(hfctmp->fifos+hfc_FIFO_B2TX_Z2+0x1f*4);
    chtmp->tx.c[1].fifo_base=(char *)(hfctmp->fifos+hfc_FIFO_B2TX_ZOFF);
    chtmp->tx.c[1].filled=0;
    chtmp->tx.z1=hfc_B_SUB_VAL;
    chtmp->tx.diff=0;

    hfc_dch_init(hfctmp);

    chtmp->initialized=0;
}

static int hfc_bch_check(struct hfc_card *hfctmp){
    struct bch *chtmp=&hfctmp->bch;
    int x,r;

    for(x=0;x<2;x++){
	chtmp->tx.c[x].filled=(chtmp->tx.z1-*chtmp->tx.c[x].z2p+hfc_B_FIFO_SIZE)%hfc_B_FIFO_SIZE;
	chtmp->rx.c[x].filled=(*chtmp->rx.c[x].z1p-chtmp->rx.z2+hfc_B_FIFO_SIZE)%hfc_B_FIFO_SIZE;
    }
    if(chtmp->fill_fifo){
	chtmp->checkcnt++;
	chtmp->checkcnt%=ZT_CHUNKSIZE;
	r=!chtmp->checkcnt;
    }else{
	x=chtmp->tx.c[0].filled-chtmp->tx.c[1].filled;
	if(abs(x-chtmp->tx.diff)>1){
	    printk(KERN_CRIT "zaphfc[%d]: tx sync changed: %d, %d\n",hfctmp->cardno,chtmp->tx.c[0].filled,chtmp->tx.c[1].filled);
	    chtmp->tx.diff=x;
	}
	r=chtmp->tx.c[0].filled<=ZT_CHUNKSIZE&&chtmp->tx.c[1].filled<=ZT_CHUNKSIZE;
    }
    return(r);
}

#define hfc_bch_inc_z(a,b) (a)=((a)-hfc_B_SUB_VAL+(b))%hfc_B_FIFO_SIZE+hfc_B_SUB_VAL

static void hfc_bch_tx(struct hfc_card *hfctmp){
    struct bch *chtmp=&hfctmp->bch;
    int x;

    for(x=0;x<2;x++)
	memcpy((void *)(chtmp->tx.c[x].fifo_base+chtmp->tx.z1),hfctmp->ztdev->chans[x].writechunk,ZT_CHUNKSIZE);
    hfc_bch_inc_z(chtmp->tx.z1,ZT_CHUNKSIZE);
    if(chtmp->fill_fifo){
	chtmp->fill_fifo=0;
    }else if(chtmp->tx.c[0].filled<=1||chtmp->tx.c[1].filled<=1){
	chtmp->fill_fifo=1;
	if(chtmp->initialized)
	    printk(KERN_CRIT "zaphfc[%d]: b channel buffer underrun: %d, %d\n",hfctmp->cardno,chtmp->tx.c[0].filled,chtmp->tx.c[1].filled);
    }
    if(!chtmp->fill_fifo)
	for(x=0;x<2;x++)*chtmp->tx.c[x].z1p=chtmp->tx.z1;
}

static void hfc_bch_rx(struct hfc_card *hfctmp){
    struct bch *chtmp=&hfctmp->bch;
    int x;

    x=chtmp->rx.c[0].filled-chtmp->rx.c[1].filled;
    if(abs(x-chtmp->rx.diff)>1){
	printk(KERN_CRIT "zaphfc[%d]: rx sync changed: %d, %d\n",hfctmp->cardno,chtmp->rx.c[0].filled,chtmp->rx.c[1].filled);
	chtmp->rx.diff=x;
    }
    if(chtmp->rx.c[0].filled>=ZT_CHUNKSIZE&&chtmp->rx.c[1].filled>=ZT_CHUNKSIZE){
	if((chtmp->rx.c[0].filled>=ZT_CHUNKSIZE*3&&chtmp->rx.c[1].filled>=ZT_CHUNKSIZE*3)||!chtmp->initialized){
	    if(chtmp->initialized)
		printk(KERN_CRIT "zaphfc[%d]: b channel buffer overflow: %d, %d\n",hfctmp->cardno,chtmp->rx.c[0].filled,chtmp->rx.c[1].filled);
	    hfc_bch_inc_z(chtmp->rx.z2,chtmp->rx.c[0].filled-chtmp->rx.c[0].filled%ZT_CHUNKSIZE-ZT_CHUNKSIZE);
	    chtmp->initialized=1;
	}
	for(x=0;x<2;x++){
	    memcpy(hfctmp->ztdev->chans[x].readchunk,(void *)(chtmp->rx.c[x].fifo_base+chtmp->rx.z2),ZT_CHUNKSIZE);
	    zt_ec_chunk(&hfctmp->ztdev->chans[x],hfctmp->ztdev->chans[x].readchunk,hfctmp->ztdev->chans[x].writechunk);
	}
	hfc_bch_inc_z(chtmp->rx.z2,ZT_CHUNKSIZE);
    }
}

/*===========================================================================*/

static void hfc_dch_tx(struct hfc_card *hfctmp){
    struct dch *chtmp=&hfctmp->dch;
    u8 tx_f2_v;
    u16 x;

    if(hfctmp->ztdev->chans[2].bytes2transmit){
	if(debug){
    	    printk(KERN_CRIT "zaphfc[%d]: card TX [ ",hfctmp->cardno);
	    for(x=0;x<hfctmp->ztdev->chans[2].bytes2transmit;x++){
		printk("%#2x ",hfctmp->dtransbuf[x]);
	    }
	    printk("] %d bytes\n",hfctmp->ztdev->chans[2].bytes2transmit);
	}
	tx_f2_v=*chtmp->tx.f2.p;
	if(!(tx_f2_v-chtmp->tx.f1.v+hfc_MAX_DFRAMES+1-1)&(hfc_MAX_DFRAMES+1-1)){
	    printk(KERN_CRIT "zaphfc[%d]: dchan tx fifo total number of frames exceeded!\n",hfctmp->cardno);
	}else{
	    if(((*(volatile u16 *)(hfctmp->fifos+hfc_FIFO_DTX_Z2+tx_f2_v*4)-chtmp->tx.f1.z1.v+hfc_D_FIFO_SIZE-1)&(hfc_D_FIFO_SIZE-1))<hfctmp->ztdev->chans[2].bytes2transmit){
		printk(KERN_CRIT "zaphfc[%d]: dchan tx fifo not enough space for frame!\n",hfctmp->cardno);
	    }else{
		chtmp->tx.f1.v=((chtmp->tx.f1.v+1)&hfc_MAX_DFRAMES)|(hfc_MAX_DFRAMES+1);
		x=min(hfctmp->ztdev->chans[2].bytes2transmit,hfc_D_FIFO_SIZE-chtmp->tx.f1.z1.v);
		memcpy(hfctmp->fifos+hfc_FIFO_DTX_ZOFF+chtmp->tx.f1.z1.v,hfctmp->ztdev->chans[2].writechunk,x);
		memcpy(hfctmp->fifos+hfc_FIFO_DTX_ZOFF,hfctmp->ztdev->chans[2].writechunk+x,hfctmp->ztdev->chans[2].bytes2transmit-x);
		*(volatile u16 *)(hfctmp->fifos+hfc_FIFO_DTX_Z2+chtmp->tx.f1.v*4)=chtmp->tx.f1.z1.v;
		chtmp->tx.f1.z1.v=(chtmp->tx.f1.z1.v+hfctmp->ztdev->chans[2].bytes2transmit+hfc_D_FIFO_SIZE)&(hfc_D_FIFO_SIZE-1);
		*(volatile u16 *)(hfctmp->fifos+hfc_FIFO_DTX_Z1+chtmp->tx.f1.v*4)=chtmp->tx.f1.z1.v;
		*chtmp->tx.f1.p=chtmp->tx.f1.v;
	    }
	}
    }
}

static void hfc_dch_rx(struct hfc_card *hfctmp){
    struct dch *chtmp=&hfctmp->dch;
    u16 size;

    hfctmp->ztdev->chans[2].bytes2receive=0;
    hfctmp->ztdev->chans[2].eofrx=0;
    if(*chtmp->rx.f1.p==chtmp->rx.f2.v){
	hfctmp->regs.int_drec=0;
    }else{
	size=((*(volatile u16 *)(hfctmp->fifos+hfc_FIFO_DRX_Z1+chtmp->rx.f2.v*4)-chtmp->rx.f2.z2.v+hfc_D_FIFO_SIZE)&(hfc_D_FIFO_SIZE-1))+1;
	if(size<4){
	    printk(KERN_CRIT "zaphfc[%d]: empty HDLC frame received.\n",hfctmp->cardno);
	}else{
	    u16 x=min(size,(u16)(hfc_D_FIFO_SIZE-chtmp->rx.f2.z2.v));
	    memcpy(hfctmp->drecbuf,hfctmp->fifos+hfc_FIFO_DRX_ZOFF+chtmp->rx.f2.z2.v,x);
	    memcpy(hfctmp->drecbuf+x,hfctmp->fifos+hfc_FIFO_DRX_ZOFF,size-x);
	    if(hfctmp->drecbuf[size-1]){
		printk(KERN_CRIT "zaphfc[%d]: received d channel frame with bad CRC.\n",hfctmp->cardno);
	    }else{
		hfctmp->ztdev->chans[2].bytes2receive=size-1;
		hfctmp->ztdev->chans[2].eofrx=1;
	    }
	}
	chtmp->rx.f2.z2.v=(chtmp->rx.f2.z2.v+size)&(hfc_D_FIFO_SIZE-1);
	chtmp->rx.f2.v=((chtmp->rx.f2.v+1)&hfc_MAX_DFRAMES)|(hfc_MAX_DFRAMES+1);
    }
}

/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

#ifdef LINUX26
static irqreturn_t hfc_interrupt(int irq, void *dev_id, struct pt_regs *regs) {
#else
static void hfc_interrupt(int irq, void *dev_id, struct pt_regs *regs) {
#endif
    struct hfc_card *hfctmp = dev_id;
    struct hfc_card *hfctmp2;
    struct zt_hfc *zthfc;
    unsigned char stat, s1, s2, l1state;
    unsigned long flags;
    unsigned long flags2=0;
    int x;

    if (!hfctmp) {
#ifdef LINUX26
		return IRQ_NONE;
#else
		return;
#endif		
    }

    if (!hfctmp->pci_io) {
	    printk(KERN_WARNING "%s: IO-mem disabled, cannot handle interrupt\n",
		   __FUNCTION__);
#ifdef LINUX26
	    return IRQ_NONE;
#else
	    return;
#endif		
    }

    spin_lock_irqsave(&hfctmp->lock, flags);
    stat = hfc_inb(hfctmp, hfc_STATUS);
    if ((stat & hfc_STATUS_ANYINT) == 0) {
        // maybe we are sharing the irq
	spin_unlock_irqrestore(&hfctmp->lock,flags);
#ifdef LINUX26
	return IRQ_NONE;
#else
	return;
#endif		
    }
    s1 = hfc_inb(hfctmp, hfc_INT_S1);
    s2 = hfc_inb(hfctmp, hfc_INT_S2); 
    if (s1 != 0) {
	if (s1 & hfc_INTS_TIMER) {
	    // timer (bit 7)
	    // printk(KERN_CRIT "timer %d %d %d.\n", stat, s1, s2);
	}
	if (s1 & hfc_INTS_L1STATE) {
	    // state machine (bit 6)
	    // printk(KERN_CRIT "zaphfc: layer 1 state machine interrupt\n");
	    zthfc = hfctmp->ztdev;
	    l1state = hfc_inb(hfctmp,hfc_STATES)  & hfc_STATES_STATE_MASK;
	    if (hfctmp->regs.nt_mode == 1) {
		if (debug) {
	    	    printk(KERN_CRIT "zaphfc: card %d layer 1 state = G%d\n", hfctmp->cardno, l1state);
		}
		switch (l1state) {
		    case 3:
			sprintf(zthfc->span.desc, "HFC-S PCI A ISDN card %d [NT] layer 1 ACTIVATED (G%d)", hfctmp->cardno, l1state);
			break;
		    default:
			sprintf(zthfc->span.desc, "HFC-S PCI A ISDN card %d [NT] layer 1 DEACTIVATED (G%d)", hfctmp->cardno, l1state);
		}
		if (l1state == 2) {
		    hfc_outb(hfctmp, hfc_STATES, hfc_STATES_ACTIVATE | hfc_STATES_DO_ACTION | hfc_STATES_NT_G2_G3);
		} else if (l1state == 3) {
		    // fix to G3 state (see specs)
		    hfc_outb(hfctmp, hfc_STATES, hfc_STATES_LOAD_STATE | 3);
		}
	    } else {
		if (debug) {
	    	    printk(KERN_CRIT "zaphfc: card %d layer 1 state = F%d\n", hfctmp->cardno, l1state);
		}
		switch (l1state) {
		    case 7:
			sprintf(zthfc->span.desc, "HFC-S PCI A ISDN card %d [TE] layer 1 ACTIVATED (F%d)", hfctmp->cardno, l1state);
			break;
		    default:
			sprintf(zthfc->span.desc, "HFC-S PCI A ISDN card %d [TE] layer 1 DEACTIVATED (F%d)", hfctmp->cardno, l1state);
		}
		if (l1state == 3) {
		    hfc_outb(hfctmp, hfc_STATES, hfc_STATES_DO_ACTION | hfc_STATES_ACTIVATE);
		}
	    }
	    
	}
	if (s1 & hfc_INTS_DREC) {
	    // D chan RX (bit 5)
	    hfctmp->regs.int_drec = 1;
	    // mr. zapata there is something for you!
	//    printk(KERN_CRIT "d chan rx\n");		    
	}
	if (s1 & hfc_INTS_B2REC) {
	    // B2 chan RX (bit 4)
	}
	if (s1 & hfc_INTS_B1REC) {
	    // B1 chan RX (bit 3)
	}
	if (s1 & hfc_INTS_DTRANS) {
	    // D chan TX (bit 2)
//	    printk(KERN_CRIT "zaphfc: dchan frame transmitted.\n");
	}
	if (s1 & hfc_INTS_B2TRANS) {
	    // B2 chan TX (bit 1)
	}
	if (s1 & hfc_INTS_B1TRANS) {
	    // B1 chan TX (bit 0)
	}
    }
    if (s2 != 0) {
	if (s2 & hfc_M2_PMESEL) {
	    // kaboom irq (bit 7)
	    // printk(KERN_CRIT "zaphfc: sync lost, pci performance too low. you might have some cpu throtteling enabled.\n");
	}
	if (s2 & hfc_M2_GCI_MON_REC) {
	    // RxR monitor channel (bit 2)
	}
	if (s2 & hfc_M2_GCI_I_CHG) {
	    // GCI I-change  (bit 1)
	}
	if((s2&hfc_M2_PROC_TRANS)&&(hfctmp->cardno==timer_card)){
	    // processing/non-processing transition  (bit 0)
	hfctmp2=hfctmp;
	hfctmp=hfc_dev_list;
	while(hfctmp){
	    if(hfctmp->active){
	    if(hfctmp!=hfctmp2)spin_lock_irqsave(&hfctmp->lock, flags2);
	    if(hfc_bch_check(hfctmp)){
		if (hfctmp->ztdev->span.flags & ZT_FLAG_RUNNING) {
		    // clear dchan buffer
	//	    memset(hfctmp->drecbuf, 0x0, sizeof(hfctmp->drecbuf));

		    hfctmp->ztdev->chans[2].bytes2transmit = 0;
		    hfctmp->ztdev->chans[2].maxbytes2transmit = hfc_D_FIFO_SIZE;

		    zt_transmit(&(hfctmp->ztdev->span));

		    hfc_bch_tx(hfctmp);
		    hfc_dch_tx(hfctmp);
		}
		hfc_bch_rx(hfctmp);
		if (hfctmp->regs.int_drec) {
		    // dchan data to read
		    hfc_dch_rx(hfctmp);
		    if (hfctmp->ztdev->chans[2].bytes2receive > 0) {
			    if (debug) {
    				printk(KERN_CRIT "zaphfc: card %d RX [ ", hfctmp->cardno);
				if (hfctmp->ztdev->chans[2].eofrx) {
				    /* dont output CRC == less user confusion */
				    for (x=0; x < hfctmp->ztdev->chans[2].bytes2receive - 2; x++) {
					printk("%#2x ", hfctmp->drecbuf[x]);
				    }
				    printk("] %d bytes\n", hfctmp->ztdev->chans[2].bytes2receive - 2);
				} else {
				    for (x=0; x < hfctmp->ztdev->chans[2].bytes2receive; x++) {
					printk("%#2x ", hfctmp->drecbuf[x]);
				    }
				    printk("..] %d bytes\n", hfctmp->ztdev->chans[2].bytes2receive);
				}
			    }
		    }
		} else {
			// hmm....ok, let zaptel receive nothing
		    hfctmp->ztdev->chans[2].bytes2receive = 0;
		}
		if (hfctmp->ztdev->span.flags & ZT_FLAG_RUNNING) {
		    zt_receive(&(hfctmp->ztdev->span));
		}
	    }
	    if(hfctmp!=hfctmp2)spin_unlock_irqrestore(&hfctmp->lock,flags2);
	    }
	    hfctmp=hfctmp->next;
	}
	hfctmp=hfctmp2;
	}
    }
    spin_unlock_irqrestore(&hfctmp->lock,flags);
#ifdef LINUX26
	return IRQ_RETVAL(1);
#endif		
}


static int zthfc_open(struct zt_chan *chan) {
    struct zt_hfc *zthfc = chan->pvt;
    struct hfc_card *hfctmp = zthfc->card;
    
    if (!hfctmp) {
    return 0;
    }
#ifndef LINUX26  
    MOD_INC_USE_COUNT;
#else
    try_module_get(THIS_MODULE);
#endif
    return 0;
}

static int zthfc_close(struct zt_chan *chan) {
    struct zt_hfc *zthfc = chan->pvt;
    struct hfc_card *hfctmp = zthfc->card;

    if (!hfctmp) {
	return 0;
    }

#ifndef LINUX26  
    MOD_DEC_USE_COUNT;
#else
    module_put(THIS_MODULE);
#endif
    return 0;
}

static int zthfc_rbsbits(struct zt_chan *chan, int bits) {
    return 0;
}

static int zthfc_ioctl(struct zt_chan *chan, unsigned int cmd, unsigned long data) {
        switch(cmd) {
        default:
                return -ENOTTY;
        }
        return 0;
}

static int zthfc_startup(struct zt_span *span) {
    struct zt_hfc *zthfc = span->pvt;
    struct hfc_card *hfctmp = zthfc->card;
    int alreadyrunning;
    
    if (hfctmp == NULL) {
	printk(KERN_INFO "zaphfc: no card for span at startup!\n");
    }
    alreadyrunning = span->flags & ZT_FLAG_RUNNING;
    
    if (alreadyrunning) return 0;

    span->chans[2].flags &= ~ZT_FLAG_HDLC;
    span->chans[2].flags |= ZT_FLAG_BRIDCHAN;
    
    span->flags |= ZT_FLAG_RUNNING;

    hfctmp->ticks = -2;	
    hfctmp->regs.fifo_en = hfc_FIFOEN_D | hfc_FIFOEN_B1 | hfc_FIFOEN_B2;
    hfc_outb(hfctmp, hfc_FIFO_EN, hfctmp->regs.fifo_en);

    hfc_bch_init(hfctmp);
    // drivers, start engines!
    hfc_outb(hfctmp, hfc_STATES, hfc_STATES_DO_ACTION | hfc_STATES_ACTIVATE);
    hfctmp->active=1;
    return 0;
}

static int zthfc_shutdown(struct zt_span *span) {
    return 0;
}

static int zthfc_maint(struct zt_span *span, int cmd) {
    return 0;
}

static int zthfc_chanconfig(struct zt_chan *chan, int sigtype) {
//    printk(KERN_CRIT "chan_config sigtype=%d\n", sigtype);
    return 0;
}

static int zthfc_spanconfig(struct zt_span *span, struct zt_lineconfig *lc) {
    span->lineconfig = lc->lineconfig;
    return 0;
}

static int zthfc_initialize(struct zt_hfc *zthfc) {
    struct hfc_card *hfctmp = zthfc->card;
    int i;

    memset(&zthfc->span, 0x0, sizeof(struct zt_span)); // you never can tell...

    sprintf(zthfc->span.name, "ZTHFC%d", hfc_dev_count + 1);
    if (hfctmp->regs.nt_mode == 1) {
	sprintf(zthfc->span.desc, "HFC-S PCI A ISDN card %d [NT]", hfc_dev_count + 1);
    } else {
	sprintf(zthfc->span.desc, "HFC-S PCI A ISDN card %d [TE]", hfc_dev_count + 1);
    }

    zthfc->span.spanconfig = zthfc_spanconfig;
    zthfc->span.chanconfig = zthfc_chanconfig;
    zthfc->span.startup = zthfc_startup;
    zthfc->span.shutdown = zthfc_shutdown;
    zthfc->span.maint = zthfc_maint;
    zthfc->span.rbsbits = zthfc_rbsbits;
    zthfc->span.open = zthfc_open;
    zthfc->span.close = zthfc_close;
    zthfc->span.ioctl = zthfc_ioctl;

    zthfc->span.chans = zthfc->chans;
    zthfc->span.channels = 3;
    zthfc->span.deflaw = ZT_LAW_ALAW;
    zthfc->span.linecompat = ZT_CONFIG_AMI | ZT_CONFIG_CCS; // <--- this is really BS
    zthfc->span.offset = 0;
    init_waitqueue_head(&zthfc->span.maintq);
    zthfc->span.pvt = zthfc;

    for (i = 0; i < zthfc->span.channels; i++) {
	memset(&(zthfc->chans[i]), 0x0, sizeof(struct zt_chan));
	sprintf(zthfc->chans[i].name, "ZTHFC%d/%d/%d", hfc_dev_count + 1,0,i + 1);
	zthfc->chans[i].pvt = zthfc;
	zthfc->chans[i].sigcap =  ZT_SIG_EM | ZT_SIG_CLEAR | ZT_SIG_FXSLS | ZT_SIG_FXSGS | ZT_SIG_FXSKS | ZT_SIG_FXOLS | ZT_SIG_FXOGS | ZT_SIG_FXOKS | ZT_SIG_CAS | ZT_SIG_SF;
	zthfc->chans[i].chanpos = i + 1; 
    }

    if (zt_register(&zthfc->span,0)) {
	printk(KERN_CRIT "unable to register zaptel device!\n");
	return -1;
    }
//    printk(KERN_CRIT "zaphfc: registered zaptel device!\n");
    return 0;
}

int hfc_findCards(int pcivendor, int pcidevice, char *vendor_name, char *card_name) {
    struct pci_dev *tmp;
    struct hfc_card *hfctmp = NULL;
    struct zt_hfc *zthfc = NULL;

    tmp = pci_find_device(pcivendor, pcidevice, multi_hfc);
    while (tmp != NULL) {
	multi_hfc = tmp;	// skip this next time.

	if (pci_enable_device(tmp)) {
	    multi_hfc = NULL;
	    return -1;
	}
	pci_set_master(tmp);

	hfctmp = vmalloc(sizeof(struct hfc_card));
	if (!hfctmp) {
	    printk(KERN_WARNING "zaphfc: unable to vmalloc!\n");
	    pci_disable_device(tmp);
	    multi_hfc = NULL;
	    return -ENOMEM;
	}
	memset(hfctmp, 0x0, sizeof(struct hfc_card));
	spin_lock_init(&hfctmp->lock);
	
	hfctmp->active=0;
	hfctmp->pcidev = tmp;
	hfctmp->pcibus = tmp->bus->number;
	hfctmp->pcidevfn = tmp->devfn; 

	if (!tmp->irq) {
	    printk(KERN_WARNING "zaphfc: no irq!\n");
	} else {
	    hfctmp->irq = tmp->irq;
	}

	hfctmp->pci_io = (char *) tmp->resource[1].start;
	if (!hfctmp->pci_io) {
	    printk(KERN_WARNING "zaphfc: no iomem!\n");
	    vfree(hfctmp);
	    pci_disable_device(tmp);
	    multi_hfc = NULL;
	    return -1;
	}

	hfctmp->fifos=(void *)__get_free_pages(GFP_KERNEL,log2(hfc_FIFO_MEM_SIZE_PAGES));
	if (!hfctmp->fifos) {
	    printk(KERN_WARNING "zaphfc: unable to __get_free_pages fifomem!\n");
	    vfree(hfctmp);
	    pci_disable_device(tmp);
	    multi_hfc = NULL;
	    return -ENOMEM;
	} else {
	    pci_write_config_dword(hfctmp->pcidev, 0x80, (u_int) virt_to_bus(hfctmp->fifos));
	    hfctmp->pci_io = ioremap((ulong) hfctmp->pci_io, 256);
	}

	if (request_irq(hfctmp->irq, &hfc_interrupt, SA_INTERRUPT | SA_SHIRQ, "zaphfc", hfctmp)) {
	    printk(KERN_WARNING "zaphfc: unable to register irq\n");
	    free_pages((unsigned long)hfctmp->fifos,log2(hfc_FIFO_MEM_SIZE_PAGES));
	    vfree(hfctmp);
	    iounmap((void *) hfctmp->pci_io);
	    pci_disable_device(tmp);
	    multi_hfc = NULL;
	    return -EIO;
	}

	printk(KERN_INFO
		       "zaphfc: %s %s configured at mem %#x fifo %#x(%#x) IRQ %d HZ %d\n",
			vendor_name, card_name,
		       (u_int) hfctmp->pci_io,
		       (u_int) hfctmp->fifos,
		       (u_int) virt_to_bus(hfctmp->fifos),
		       hfctmp->irq, HZ); 
	pci_write_config_word(hfctmp->pcidev, PCI_COMMAND, PCI_COMMAND_MEMORY);	// enable memio
	hfctmp->regs.int_m1 = 0;	// no ints
	hfctmp->regs.int_m2 = 0;	// not at all
	hfc_outb(hfctmp,hfc_INT_M1,hfctmp->regs.int_m1);
	hfc_outb(hfctmp,hfc_INT_M2,hfctmp->regs.int_m2);

	if ((modes & (1 << hfc_dev_count)) != 0) {
	    printk(KERN_INFO "zaphfc: Card %d configured for NT mode\n",hfc_dev_count);
	    hfctmp->regs.nt_mode = 1;
	} else {
	    printk(KERN_INFO "zaphfc: Card %d configured for TE mode\n",hfc_dev_count);
	    hfctmp->regs.nt_mode = 0;
	}

	if(sync_slave&(1<<hfc_dev_count)){
	    printk(KERN_INFO "zaphfc: Card %d configured for slave mode\n",hfc_dev_count);
	    hfctmp->regs.mst_mode=hfc_MST_MODE_SLAVE|hfc_MST_MODE_F0_LONG_DURATION;
	    hfctmp->regs.mst_emod=hfc_MST_EMOD_SLOW_CLOCK_ADJ;
	}else{
	    printk(KERN_INFO "zaphfc: Card %d configured for master mode\n",hfc_dev_count);
	    hfctmp->regs.mst_mode=hfc_MST_MODE_MASTER|hfc_MST_MODE_F0_LONG_DURATION;
	    hfctmp->regs.mst_emod=0;
	}

	zthfc = vmalloc(sizeof(struct zt_hfc));
	if (!zthfc) {
	    printk(KERN_CRIT "zaphfc: unable to vmalloc!\n");
	    hfc_shutdownCard(hfctmp);
	    vfree(hfctmp);
	    multi_hfc = NULL;
	    return -ENOMEM;
	}
	memset(zthfc, 0x0, sizeof(struct zt_hfc));

	zthfc->card = hfctmp;
	zthfc_initialize(zthfc);
	hfctmp->ztdev = zthfc;

	memset(hfctmp->drecbuf, 0x0, sizeof(hfctmp->drecbuf));
	hfctmp->ztdev->chans[2].readchunk = hfctmp->drecbuf;

	memset(hfctmp->dtransbuf, 0x0, sizeof(hfctmp->dtransbuf));
	hfctmp->ztdev->chans[2].writechunk = hfctmp->dtransbuf;

	memset(hfctmp->brecbuf[0], 0x0, sizeof(hfctmp->brecbuf[0]));
	hfctmp->ztdev->chans[0].readchunk = hfctmp->brecbuf[0];
	memset(hfctmp->btransbuf[0], 0x0, sizeof(hfctmp->btransbuf[0]));
	hfctmp->ztdev->chans[0].writechunk = hfctmp->btransbuf[0];

	memset(hfctmp->brecbuf[1], 0x0, sizeof(hfctmp->brecbuf[1]));
	hfctmp->ztdev->chans[1].readchunk = hfctmp->brecbuf[1];
	memset(hfctmp->btransbuf[1], 0x0, sizeof(hfctmp->btransbuf[1]));
	hfctmp->ztdev->chans[1].writechunk = hfctmp->btransbuf[1];

	hfc_registerCard(hfctmp);
	hfc_resetCard(hfctmp);
	tmp = pci_find_device(pcivendor, pcidevice, multi_hfc);
    }
    return 0;
}

int init_module(void) {
    int i = 0;
    while (id_list[i].vendor_id) {
	multi_hfc = NULL;
	hfc_findCards(id_list[i].vendor_id, id_list[i].device_id, id_list[i].vendor_name, id_list[i].card_name);
	i++;
    }
    printk(KERN_INFO "zaphfc: %d hfc-pci card(s) in this box.\n", hfc_dev_count);
    return 0;
}

void cleanup_module(void) {
    struct hfc_card *tmpcard;

    printk(KERN_INFO "zaphfc: stop\n");
//    spin_lock(&registerlock);
    tmpcard=hfc_dev_list;
    while(tmpcard){
	hfc_shutdownCard1(tmpcard);
	tmpcard=tmpcard->next;
    }
    while (hfc_dev_list != NULL) {
	if (hfc_dev_list == NULL) break;
	hfc_shutdownCard2(hfc_dev_list);
	tmpcard = hfc_dev_list;
	hfc_dev_list = hfc_dev_list->next;
	if (tmpcard != NULL) {
	    vfree(tmpcard);
	    tmpcard = NULL;
	    printk(KERN_INFO "zaphfc: freed one card.\n");
	}
    }
//    spin_unlock(&registerlock);
}
#endif


MODULE_DESCRIPTION("HFC-S PCI A Zaptel Driver");
MODULE_AUTHOR("Klaus-Peter Junghanns <kpj@junghanns.net>");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif	
MODULE_PARM(modes,"i");
MODULE_PARM(debug,"i");
MODULE_PARM(sync_slave,"i");
MODULE_PARM(timer_card,"i");

/* vim:set sw=4: */

